use std::{
    collections::HashMap,
    ops::Deref,
    path::{Path, PathBuf},
};

use anyhow::Context;
use tokio::io::AsyncWriteExt;
use tokio_stream::{wrappers::ReadDirStream, StreamExt};

pub async fn process() -> anyhow::Result<()> {
    process_opts(ProcessOpts::default()).await
}

pub struct ProcessOpts {
    pub path: PathBuf,
    pub output: PathBuf,
}

impl Default for ProcessOpts {
    fn default() -> Self {
        Self {
            path: std::env::current_dir().expect("to be able to get current dir"),
            output: std::env::current_dir()
                .expect("to be able to get current dir")
                .join("cuddle-clusters")
                .join("k8s"),
        }
    }
}

pub async fn process_opts(opts: ProcessOpts) -> anyhow::Result<()> {
    let path = opts.path.canonicalize().context("failed to find folder")?;

    let cuddle_path = path.join("cuddle.yaml");
    let template = path.join("templates").join("clusters");

    tracing::debug!(
        "searching for templates in: {} with cuddle: {}",
        template.display(),
        cuddle_path.display()
    );

    let clusters = read_cuddle_section(&cuddle_path).await?;
    tracing::debug!("found clusters: {:?}", clusters);

    let template_files = load_template_files(&template).await?;
    tracing::debug!("found files: {:?}", template_files);

    tokio::fs::remove_dir_all(&opts.output).await?;
    tokio::fs::create_dir_all(&opts.output).await?;

    process_templates(&clusters, &template_files, &opts.output).await?;

    Ok(())
}

#[derive(serde::Deserialize, Default, Debug, Clone)]
struct CuddleClusters(HashMap<String, serde_yaml::Value>);

impl Deref for CuddleClusters {
    type Target = HashMap<String, serde_yaml::Value>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

async fn read_cuddle_section(path: &Path) -> anyhow::Result<CuddleClusters> {
    let cuddle_file = tokio::fs::read(path)
        .await
        .context(format!("failed to read: {}", path.display()))?;

    let value: serde_yaml::Value = serde_yaml::from_slice(&cuddle_file)?;

    let mut cuddle_clusters = CuddleClusters::default();
    if let Some(clusters) = value.get("cuddle/clusters") {
        if let Some(mapping) = clusters.as_mapping() {
            for (key, value) in mapping.iter() {
                if let Some(key) = key.as_str() {
                    cuddle_clusters.0.insert(key.to_string(), value.clone());
                }
            }
        }
    }

    Ok(cuddle_clusters)
}

#[derive(Debug, Clone)]
struct TemplateFiles {
    templates: Vec<PathBuf>,
    raw: Vec<PathBuf>,
}

async fn load_template_files(path: &Path) -> anyhow::Result<TemplateFiles> {
    Ok(TemplateFiles {
        templates: read_dir(path)
            .await?
            .into_iter()
            .filter(|i| !i.ends_with(".jinja2"))
            .collect(),
        raw: read_dir(&path.join("raw")).await.unwrap_or_default(),
    })
}

async fn read_dir(path: &Path) -> anyhow::Result<Vec<PathBuf>> {
    let template_dir = tokio::fs::read_dir(path).await?;
    let mut template_dir_stream = ReadDirStream::new(template_dir);

    let mut paths = Vec::new();
    while let Some(entry) = template_dir_stream.next().await {
        let entry = entry?;

        if entry.metadata().await?.is_file() {
            paths.push(entry.path());
        }
    }

    Ok(paths)
}

async fn process_templates(
    clusters: &CuddleClusters,
    template_files: &TemplateFiles,
    dest: &Path,
) -> anyhow::Result<()> {
    for (cluster_name, _value) in clusters.iter() {
        process_cluster(cluster_name, template_files, &dest.join(cluster_name)).await?;
    }

    Ok(())
}

async fn process_cluster(
    cluster_name: &str,
    template_files: &TemplateFiles,
    dest: &Path,
) -> anyhow::Result<()> {
    for template_file in &template_files.templates {
        process_template_file(cluster_name, template_file, dest).await?;
    }

    for raw_file in &template_files.raw {
        process_raw_file(cluster_name, raw_file, dest).await?;
    }

    Ok(())
}

async fn process_template_file(
    cluster_name: &str,
    template_file: &PathBuf,
    dest: &Path,
) -> anyhow::Result<()> {
    // TODO: use mini jinja
    let file = tokio::fs::read_to_string(template_file)
        .await
        .context(format!("failed to find file: {}", template_file.display()))?;

    if !dest.exists() {
        tokio::fs::create_dir_all(dest).await?;
    }

    let file_name = template_file
        .file_stem()
        .ok_or(anyhow::anyhow!("file didn't have a jinja2 format"))?;

    let mut dest_file = tokio::fs::File::create(dest.join(file_name)).await?;
    dest_file.write_all(file.as_bytes()).await?;

    Ok(())
}

async fn process_raw_file(
    _cluster_name: &str,
    raw_file: &PathBuf,
    dest: &Path,
) -> anyhow::Result<()> {
    let file = tokio::fs::read_to_string(raw_file)
        .await
        .context(format!("failed to find file: {}", raw_file.display()))?;

    if !dest.exists() {
        tokio::fs::create_dir_all(dest).await?;
    }

    let file_name = raw_file
        .file_name()
        .ok_or(anyhow::anyhow!("file didn't have a file name"))?;

    let mut dest_file = tokio::fs::File::create(dest.join(file_name)).await?;
    dest_file.write_all(file.as_bytes()).await?;

    Ok(())
}
